const { Container, DependencyInjector } = require('flitter-di')
const { test_connect } = require('../common')
const { expect } = require('chai')
const { MongoClient, ObjectId } = require('mongodb')
const uuid = require('uuid/v4')

describe('basic model storage and retrieval', function() {
    const Connection = require('../../src/services/Connection')
    const Model = require('../../src/model/Model')
    const di = new DependencyInjector(new Container({ scaffold: Connection }))
    class TestModel extends Model {
        static collection = uuid()

        static get schema() {
            return {
                name: String,
                number: Number,
            }
        }
    }
    di.make(TestModel)

    it('should allow opening connections with the scaffold service', async function() {
        const sc = di.service('scaffold')
        expect(sc).to.be.an.instanceof(Connection)

        const client = await sc.open(test_connect)
        expect(client).to.be.an.instanceof(MongoClient)
    })

    it('should allow saving of models', async function() {
        const test = new TestModel({name: 'a test name', number: 3.141})
        const ret = await test.save()
        expect(ret).to.be.eql(test)
        expect(ret._id).to.be.an.instanceof(ObjectId)
    })

    it('should allow retrieval of models', async function() {
        const tests = await TestModel.find()
        expect(tests.length).to.be.equal(1)
        expect(tests[0]).to.be.an.instanceof(TestModel)
        expect(tests[0].name).to.be.equal('a test name')
        expect(tests[0].number).to.be.equal(3.141)

        const test = await TestModel.findOne({name: 'a test name'})
        expect(test).to.be.an.instanceof(TestModel)
        expect(test).to.be.eql(tests[0])

        const no_tests = await TestModel.find({name: 'not a real name'})
        expect(no_tests.length).to.be.equal(0)

        const no_test = await TestModel.findOne({number: 9934})
        expect(no_test).to.be.undefined
    })

    it('should allow updating of models', async function() {
        const name = uuid()
        const test = new TestModel({name})
        await test.save()

        test.number = 993425
        await test.save()

        expect((await TestModel.findOne({name})).number).to.be.equal(test.number)
    })

    it('should allow lookups by ID', async function() {
        const name = uuid()
        const test = new TestModel({name})
        await test.save()

        const lookup_1 = await TestModel.findById(test.id)
        const lookup_2 = await TestModel.findById(test._id)

        expect(lookup_1).to.be.eql(test)
        expect(lookup_2).to.be.eql(test)
    })

    after(function() {
        di.service('scaffold').close()
    })
})
