const { expect } = require('chai')
const Filter = require('../../src/filter/Filter')
const { DependencyInjector, Container } = require('flitter-di')
const uuid = require('uuid')
const Connection = require('../../src/services/Connection')
const { test_connect } = require('../common')

describe('the filter builder classes', function() {
    const Connection = require('../../src/services/Connection')
    const Model = require('../../src/model/Model')
    const di = new DependencyInjector(new Container({ scaffold: Connection }))
    class TestModel extends Model {
        static collection = uuid()

        static get schema() {
            return {
                name: String,
                number: Number,
            }
        }
    }
    di.make(TestModel)

    before(async function() {
        await di.service('scaffold').open(test_connect)
        for ( let i = 1; i < 50; i++ ) {
            await (new TestModel({number: i, name: `number ${i}`})).save()
        }
    })

    after(function() {
        di.service('scaffold').close()
    })

    it('should allow filter objects', async function() {
        const filter_1 = await TestModel.filter()
        expect(filter_1).to.be.an.instanceof(Filter)
    })

    it('should allow equality filtering', async function() {
        const filter_2 = (await TestModel.filter()).equal('number', 24)
        const result_2 = await filter_2.end().find()
        expect(result_2.length).to.be.equal(1)
        expect(result_2[0].name).to.be.equal('number 24')
    })

    it('should allow limit filtering', async function() {
        const filter_3 = (await TestModel.filter())
            .greater_than('number', 5)
            .less_than('number', 11)
        const result_3 = await filter_3.end().find()
        expect(result_3.length).to.be.equal(5)
        const values_3 = result_3.map(x => x.number)
        expect(values_3).to.have.all.members([6, 7, 8, 9, 10])
    })

    it('should allow inclusive limit filtering', async function() {
        const filter_4 = (await TestModel.filter())
            .greater_than_equal('number', 5)
            .less_than_equal('number', 11)
        const result_4 = await filter_4.end().find()
        expect(result_4.length).to.be.equal(7)
        const values_4 = result_4.map(x => x.number)
        expect(values_4).to.have.all.members([5, 6, 7, 8, 9, 10, 11])
    })

    it('should allow set sorting', async function() {
        const filter_5 = (await TestModel.filter())
            .in('number', [3, 6, 8, 9, 44, 11])
        const result_5 = await filter_5.end().find()
        expect(result_5.length).to.be.equal(6)
        const values_5 = result_5.map(x => x.number)
        expect(values_5).to.have.all.members([3, 6, 8, 9, 44, 11])
    })
})
