const { Container, DependencyInjector } = require('flitter-di')
const { test_connect } = require('../common')
const { expect } = require('chai')
const { MongoClient, ObjectId } = require('mongodb')
const uuid = require('uuid/v4')

describe('transparent model scoping', function() {
    const Connection = require('../../src/services/Connection')
    const Model = require('../../src/model/Model')
    const Scope = require('../../src/model/Scope')
    const di = new DependencyInjector(new Container({ scaffold: Connection }))
    class TestModel extends Model {
        static collection = uuid()

        static get schema() {
            return {
                name: String,
                number: Number,
            }
        }
    }
    di.make(TestModel)

    before(async function() {
        await di.service('scaffold').open(test_connect)
        for ( let i = 1; i < 50; i++ ) {
            await (new TestModel({number: i, name: `number ${i}`})).save()
        }
    })

    after(async function() {
        await di.service('scaffold').close()
    })

    it('should limit models to their scopes', async function() {
        class TestScope extends Scope {
            async filter(to_filter) {
                return to_filter.greater_than('number', 20).less_than('number', 31)
            }
        }

        TestModel.scopes.push(new TestScope)
        const results = await TestModel.find()
        expect(results.length).to.be.equal(10)
        TestModel.scopes = []
    })

    it('should allow scopes to modify the records before save', async function() {
        class TestScope extends Scope {
            async save(record) {
                record.name = `number ${record.number}`
                return record
            }
        }

        TestModel.scopes.push(new TestScope)
        const new_model = new TestModel({number: 5432})
        await new_model.save()
        expect(new_model.name).to.be.equal('number 5432')
        TestModel.scopes = []
    })
})
