const { expect } = require('chai')
const sinon = require('sinon')

const CursorBuilder = require('../../src/model/CursorBuilder')

describe('the cursor builder', function() {
    let cb
    let fake_cursor
    beforeEach(function() {
        cb = new CursorBuilder()
        fake_cursor = {
            filter: sinon.spy()
        }
    })

    it('should be a functional class', function() {
        expect(CursorBuilder).to.be.a('function')
        expect(new CursorBuilder()).to.be.an.instanceof(CursorBuilder)
    })

    it('should have a layers property', function() {
        expect(cb.layers).to.be.an('array')
    })

    describe('the filter method', function() {
        let fake_filter
        let fake_filter_item
        beforeEach(function() {
            fake_filter_item = {}
            fake_filter = {
                write: () => fake_filter_item
            }
        })
        it('should be a function', function() {
            expect(cb.filter).to.be.a('function')
        })

        it('should expect one argument', function() {
            expect(cb.filter.length).to.be.equal(1)
        })

        it('should add a function to the layers', function() {
            cb.filter(fake_filter)
            expect(cb.layers.length).to.be.equal(1)
            expect(cb.layers[0]).to.be.a('function')
        })

        it('should apply the filter to the cursor', function() {
            cb.filter(fake_filter)
            const filter_fn = cb.layers[0]
            const fake_cursor = {
                filter: sinon.spy()
            }

            filter_fn(fake_cursor)
            expect(fake_cursor.filter.calledOnce).to.be.true
            expect(fake_cursor.filter.getCall(0).args[0]).to.be.eql(fake_filter_item)
        })
    })

    describe('the limit method', function() {
        it('should be a function', function() {
            expect(cb.limit).to.be.a('function')
        })

        it('should expect one argument', function() {
            expect(cb.limit.length).to.be.equal(1)
        })

        it('should add a function to the layers', function() {
            cb.limit(7)
            expect(cb.layers.length).to.be.equal(1)
            expect(cb.layers[0]).to.be.a('function')
        })

        it('should apply the limit to the cursor', function() {
            cb.limit(33)
            const limit_fn = cb.layers[0]
            const fake_cursor = {
                limit: sinon.spy()
            }

            limit_fn(fake_cursor)
            expect(fake_cursor.limit.calledOnce).to.be.true
            expect(fake_cursor.limit.getCall(0).args[0]).to.be.equal(33)
        })
    })

    describe('the sort method', function() {
        it('should be a function', function() {
            expect(cb.sort).to.be.a('function')
        })

        it('should expect one argument', function() {
            expect(cb.sort.length).to.be.equal(1)
        })

        it('should add a function to the layers', function() {
            cb.sort(['-fake_field'])
            expect(cb.layers.length).to.be.equal(1)
            expect(cb.layers[0]).to.be.a('function')
        })

        it('build a mongodb sort object and apply it to the cursor', function() {
            const sorts = ['-field_a', '+field_b', 'field_c']
            const expected_sort_obj = { field_a: -1, field_b: 1, field_c: 1 }

            cb.sort(sorts)
            const sort_fn = cb.layers[0]
            const fake_cursor = {
                sort: sinon.spy()
            }

            sort_fn(fake_cursor)
            expect(fake_cursor.sort.calledOnce).to.be.true
            expect(fake_cursor.sort.getCall(0).args[0]).to.be.eql(expected_sort_obj)
        })
    })

    describe('the apply method', function() {
        it('should be a function', function() {
            expect(cb.apply).to.be.a('function')
        })

        it('should expect one argument', function() {
            expect(cb.apply.length).to.be.equal(1)
        })

        it('should apply the layers to the cursor', function() {
            const layer_a = sinon.spy()
            const layer_b = sinon.spy()
            cb.layers.push(layer_a, layer_b)

            const fake_cursor = {}
            cb.apply(fake_cursor)

            expect(layer_a.calledOnce).to.be.true
            expect(layer_b.calledOnce).to.be.true
            expect(layer_a.getCall(0).args[0]).to.be.equal(fake_cursor)
            expect(layer_b.getCall(0).args[0]).to.be.equal(fake_cursor)
        })
    })
})
