const { expect } = require('chai')
const sinon = require('sinon')
const common = require('../common')
const { Injectable, DependencyInjector, Container } = require('flitter-di')
const Connection = require('../../src/services/Connection')
const { Cursor, ObjectId, Collection } = require('mongodb')

const Model = require('../../src/model/Model')
const Schema = require('../../src/schema/Schema')
const LimitProxy = require('../../src/proxy/model/LimitProxy')
const SortProxy = require('../../src/proxy/model/SortProxy')
const Filter = require('../../src/filter/Filter')

describe('the base model class', function() {
    let di
    let scaffold
    let Test
    beforeEach(async function() {
        di = new DependencyInjector(new Container({ scaffold: Connection }))
        scaffold = di.service('scaffold')
        await scaffold.open(common.test_connect)

        class TestModel extends Model {
            static get schema() {
                return {
                    num: Number,
                    name: String
                }
            }
        }
        Test = TestModel
        di.make(Test)
    })
    afterEach(function() {
        scaffold.close()
    })

    it('should be a functional class', function() {
        expect(Model).to.be.a('function')
        expect(Model.prototype).to.be.an.instanceof(Injectable)
        expect(new Model).to.be.an.instanceof(Model)
    })

    it('should request the scaffold service', function() {
        expect(Model.services).to.have.all.members(['scaffold'])
    })

    it('should have a static schema getter', function() {
        expect(Model.schema).to.exist
    })

    it('should have a static scopes array', function() {
        expect(Model.scopes).to.be.an('array')
        expect(Model.scopes.length).to.be.equal(0)
    })

    it('should have a __static schema getter', function() {
        expect(Model.__schema).to.exist
        expect(Model.__schema).to.be.an.instanceof(Schema)
    })

    it('should have a static collection getter', function() {
        expect(Model.collection).to.exist
        expect(Model.collection).to.be.false
    })

    describe('the static __name getter', function() {
        it('should return the collection name, if set', function() {
            class TestModel extends Model {
                static collection = 'a test name'
            }
            expect(TestModel.__name).to.be.equal('a test name')
        })

        it('should return the class name otherwise', function() {
            class TestModel2 extends Model {}
            expect(TestModel2.__name).to.be.equal('TestModel2')
        })
    })

    describe('the static cursor method', function() {
        it('should be a function', function() {
            expect(Test.cursor).to.be.a('function')
        })

        it('should return a promise', function() {
            expect(Test.cursor()).to.be.a('promise')
        })

        it('should resolve to a mongodb cursor', async function() {
            const c = await Test.cursor()
            expect(c).to.be.an.instanceof(Cursor)
            expect(c.cursorState.cmd.find.split('.').reverse()[0]).to.be.equal('TestModel')
        })

        it('should apply the passed in filters to the cursor', async function() {
            const expect_filters = { $and: [{test_field: 'test_val'}] }
            const c = await Test.cursor({test_field: 'test_val'})
            expect(c.cursorState.cmd.query).to.be.eql(expect_filters)
        })
    })

    describe('the static from_cursor method', function() {
        let fake_cursor
        let fake_arrays
        beforeEach(function() {
            fake_arrays = [
                { num: 3, name: 'test a' },
                { num: 77, name: 'testing 123' },
            ]
            fake_cursor = {
                toArray() {
                    return Promise.resolve(fake_arrays)
                }
            }
        })

        it('should be a function', function() {
            expect(Test.from_cursor).to.be.a('function')
        })

        it('should expect one argument', function() {
            expect(Test.from_cursor.length).to.be.equal(1)
        })

        it('should return an array of instances from the cursor data', async function() {
            const ret = await Test.from_cursor(fake_cursor)
            expect(ret.length).to.be.equal(2)
            expect(ret[0].num).to.be.equal(fake_arrays[0].num)
            expect(ret[0].name).to.be.equal(fake_arrays[0].name)
            expect(ret[1].num).to.be.equal(fake_arrays[1].num)
            expect(ret[1].name).to.be.equal(fake_arrays[1].name)
            expect(ret[0]).to.be.an.instanceof(Test)
            expect(ret[1]).to.be.an.instanceof(Test)
        })
    })

    describe('the static __collection method', function() {
        it('should be a function', function() {
            expect(Test.__collection).to.be.a('function')
        })

        it('should return a mongodb collection', function() {
            const coll = Test.__collection()
            expect(coll).to.be.an.instanceof(Collection)
            expect(coll.collectionName).to.be.equal('TestModel')
        })
    })

    describe('the static find method', function() {
        function check_range(input, min = 1, max = 100) {
            const should_have = (max - min) + 1
            if ( input.length !== should_have ) return false
            const does_have = Array(max+1).fill(false)
            input.forEach((rec) => {
                if (
                    rec.num <= max
                    && rec.name === 'number '+rec.num
                ) does_have[rec.num] = true
            })

            return does_have.slice(min).every(Boolean)
        }

        beforeEach(async function() {
            for ( let i = 1; i <= 100; i++ ) {
                await scaffold.collection('TestModel').insertOne({
                    num: i,
                    name: 'number '+i
                })
            }
        })

        afterEach(async function() {
            await scaffold.collection('TestModel').deleteMany()
        })

        it('should be a function', function() {
            expect(Test.find).to.be.a('function')
        })

        it('should return a promise', async function() {
            const p = Test.find()
            expect(p).to.be.a('promise')
            await p
        })

        it('should resolve to instances of the model', async function() {
            const recs = await Test.find()
            expect(recs).to.be.an('array')
            expect(recs[0]).to.be.an.instanceof(Test)
        })

        it('should return all instances of the model by default', async function() {
            const recs = await Test.find()
            expect(check_range(recs)).to.be.true
        })

        it('should respect passed in filters', async function() {
            let recs = await Test.find({num: 5})
            expect(recs.length).to.be.equal(1)
            expect(check_range(recs, 5, 5)).to.be.true

            recs = await Test.find({num: {$gt: 5}})
            expect(check_range(recs, 6)).to.be.true

            recs = await Test.find({num: {$in: [44, 45, 46, 47, 48, 49]}})
            expect(check_range(recs, 44, 49)).to.be.true
        })

        // TODO it should respect scopes applied to the model
    })

    describe('the static findById method', function() {
        let oid
        beforeEach(async function() {
            oid = new ObjectId
            await scaffold.collection('TestModel').insertOne({
                _id: oid,
                num: 3.141,
                name: 'pi'
            })
        })

        afterEach(async function() {
            await scaffold.collection('TestModel').deleteMany()
        })

        it('should be a function', function() {
            expect(Test.findById).to.be.a('function')
            expect(Test.findById.length).to.be.equal(1)
        })

        it('should return a promise', async function() {
            const p = Test.findById(new ObjectId)
            expect(p).to.be.a('promise')
            await p
        })

        it('should resolve to a single instance of the model', async function() {
            const m = await Test.findById(oid)
            expect(m).to.be.an.instanceof(Test)
            expect(m.name).to.be.equal('pi')
            expect(m.num).to.be.equal(3.141)
            expect(`${m._id}`).to.be.equal(`${oid}`)
        })

        it('should resolve undefined if the object id is not found', async function() {
            const m = await Test.findById(new ObjectId)
            expect(m).to.be.undefined
        })

        it('should work with string form object ids', async function() {
            const m = await Test.findById(`${oid}`)
            expect(m).to.be.an.instanceof(Test)
            expect(m.name).to.be.equal('pi')
            expect(m.num).to.be.equal(3.141)
            expect(`${m._id}`).to.be.equal(`${oid}`)
        })
    })

    describe('the static findOne method', function() {
        beforeEach(async function() {
            for ( let i = 1; i <= 100; i++ ) {
                await scaffold.collection('TestModel').insertOne({
                    num: i,
                    name: 'number '+i
                })
            }
        })

        afterEach(async function() {
            await scaffold.collection('TestModel').deleteMany()
        })

        it('should be a function', function() {
            expect(Test.findOne).to.be.a('function')
        })

        it('should return a promise', async function() {
            const p = Test.findOne()
            expect(p).to.be.a('promise')
            await p
        })

        it('should resolve to instances of the model', async function() {
            const rec = await Test.findOne()
            expect(rec).to.be.an.instanceof(Test)
        })

        it('should respect passed in filters', async function() {
            let rec = await Test.findOne({num: 5})
            expect(rec.num).to.be.equal(5)
            expect(rec.name).to.be.equal('number '+5)

            rec = await Test.findOne({num: {$gt: 5, $lt: 7}})
            expect(rec.num).to.be.equal(6)
            expect(rec.name).to.be.equal('number '+6)

            rec = await Test.findOne({num: {$in: [8]}})
            expect(rec.num).to.be.equal(8)
            expect(rec.name).to.be.equal('number '+8)
        })

        // TODO it should respect scopes applied to the model
    })

    describe('the static deleteOne method', function() {
        beforeEach(async function() {
            for ( let i = 1; i <= 100; i++ ) {
                await scaffold.collection('TestModel').insertOne({
                    num: i,
                    name: 'number '+i
                })
            }
        })

        afterEach(async function() {
            await scaffold.collection('TestModel').deleteMany()
        })

        it('should be a function', function() {
            expect(Test.deleteOne).to.be.a('function')
        })

        it('should return a promise', async function() {
            const p = Test.deleteOne()
            expect(p).to.be.a('promise')
            await p
        })

        it('should resolve undefined', async function() {
            expect(await Test.deleteOne()).to.be.undefined
        })

        it('should delete a single model record', async function() {
            const before = await scaffold.collection('TestModel').find().toArray()
            await Test.deleteOne()
            const after = await scaffold.collection('TestModel').find().toArray()
            expect(before.length).to.be.equal(after.length + 1)
        })

        it('should respect passed in filters', async function() {
            let before = (await scaffold.collection('TestModel')
                .find({num: 22})
                .limit(1)
                .toArray())[0]
            expect(before).to.exist
            expect(before.num).to.be.equal(22)
            expect(before.name).to.be.equal('number 22')
            await Test.deleteOne({num: 22})

            let after = (await scaffold.collection('TestModel')
                .find({num: 22})
                .limit(1)
                .toArray())[0]
            expect(after).not.to.exist
        })

        // TODO it should respect scopes applied to the model
    })

    describe('the static deleteMany method', function() {
        beforeEach(async function() {
            for ( let i = 1; i <= 100; i++ ) {
                await scaffold.collection('TestModel').insertOne({
                    num: i,
                    name: 'number '+i
                })
            }
        })

        afterEach(async function() {
            await scaffold.collection('TestModel').deleteMany()
        })

        it('should be a function', function() {
            expect(Test.deleteMany).to.be.a('function')
        })

        it('should return a promise', async function() {
            const p = Test.deleteMany()
            expect(p).to.be.a('promise')
            await p
        })

        it('should resolve undefined', async function() {
            expect(await Test.deleteMany()).to.be.undefined
        })

        it('should delete multiple instances of the model', async function() {
            const before_recs = await scaffold.collection('TestModel').find().toArray()
            await Test.deleteMany()
            const after_recs = await scaffold.collection('TestModel').find().toArray()
            expect(before_recs.length).to.not.be.equal(0)
            expect(after_recs.length).to.be.equal(0)
        })

        it('should respect the passed in filters', async function() {
            const before_recs = await scaffold.collection('TestModel').find().toArray()
            await Test.deleteMany({num: {$in: [22,55,3,9,17,64]}})
            const after_recs = await scaffold.collection('TestModel').find().toArray()
            expect(after_recs.length).to.be.equal(before_recs.length - 6)
            expect(after_recs.some(x => x.num === 22)).to.be.false
            expect(after_recs.some(x => x.num === 55)).to.be.false
            expect(after_recs.some(x => x.num === 3)).to.be.false
            expect(after_recs.some(x => x.num === 9)).to.be.false
            expect(after_recs.some(x => x.num === 17)).to.be.false
            expect(after_recs.some(x => x.num === 64)).to.be.false
        })

        // TODO it should respect scopes applied to the model
    })

    describe('the static limit method', function() {
        it('should be a function', function() {
            expect(Test.limit).to.be.a('function')
            expect(Test.limit.length).to.be.equal(1)
        })

        it('should return a LimitProxy', function() {
            expect(Test.limit(10)).to.be.an.instanceof(LimitProxy)
        })

        it('should set the proxy limit to the passed in value', function() {
            expect(Test.limit(37)._limit).to.be.equal(37)
        })
    })

    describe('the static sort method', function() {
        it('should be a function', function() {
            expect(Test.sort).to.be.a('function')
            expect(Test.sort.length).to.be.equal(0)
        })

        it('should return a SortProxy', function() {
            expect(Test.sort('-num', 'name')).to.be.an.instanceof(SortProxy)
        })

        it('should set the proxy limit to the passed in value', function() {
            expect(Test.sort('-num', 'name')._sorts).to.be.eql(['-num', 'name'])
        })
    })

    describe('the static count method', function() {
        beforeEach(async function() {
            for ( let i = 1; i <= 100; i++ ) {
                await scaffold.collection('TestModel').insertOne({
                    num: i,
                    name: 'number '+i
                })
            }
        })

        afterEach(async function() {
            await scaffold.collection('TestModel').deleteMany()
        })

        it('should be a function', function() {
            expect(Test.count).to.be.a('function')
        })

        it('should return a promise', async function() {
            const p = Test.count()
            expect(p).to.be.a('promise')
            await p
        })

        it('should return the count of the matched records', async function() {
            expect(await Test.count()).to.be.equal(100)
        })

        it('should respect the passed in filters', async function() {
            expect(await Test.count({num: 3})).to.be.equal(1)
            expect(await Test.count({num: {$gt: 70, $lt: 99}})).to.be.equal(28)
        })

        // TODO it should respect the scopes applied to the model
    })

    describe('the static filter method', function() {
        it('should be a function', function() {
            expect(Test.filter).to.be.a('function')
        })

        it('should return a promise', async function() {
            const p = Test.filter()
            expect(p).to.be.a('promise')
            await p
        })

        it('should resolve to a Filter instance', async function() {
            const filter = await Test.filter()
            expect(filter).to.be.an.instanceof(Filter)
            expect(filter._model).to.be.equal(Test)
        })
    })

    // --- Non-Statics
    describe('the constructor', function() {
        it('should be a function', function(){
            expect(Test).to.be.a('function')
        })

        it('should return an instance of the model', function() {
            expect(new Test).to.be.an.instanceof(Test)
        })

        it('should load passed in data into the model, by the schema', function() {
            const t = new Test({ name: 'test name', num: 44 })
            expect(t.name).to.be.equal('test name')
            expect(t.num).to.be.equal(44)
        })

        it('should ignore fields not in the schema', function() {
            const t = new Test({ name: 'ignore test', num: 24, foobar: 3.14 })
            expect(t.foobar).to.not.exist
        })
    })

    describe('the id getter', function() {
        it('should return the string form of the _id if defined', function(){
            const t = new Test
            t._id = new ObjectId
            expect(t.id).to.be.equal(`${t._id}`)
        })

        it('should return nothing if there is no _id', function() {
            expect((new Test).id).to.be.undefined
        })
    })

    // TODO describe the save method

    describe('the set method', function() {
        let t
        beforeEach(function() {
            t = new Test
            t.save = sinon.spy()
        })

        it('should be a function', function() {
            expect(t.set).to.be.a('function')
        })

        it('should expect 2 arguments', function() {
            expect(t.set.length).to.be.equal(2)
        })

        it('should set the specified field to the value', function() {
            t.set('name', 'this is a test name!')
            expect(t.name).to.be.equal('this is a test name!')
        })

        it('should call the save method afterward', function() {
            t.set('number', 23)
            expect(t.save.calledOnce).to.be.true
        })
    })

    describe('the delete method', function() {
        it('should be a function', function() {
            expect((new Test).delete).to.be.a('function')
        })

        it('should return a promise', async function() {
            const t = new Test
            const p = t.delete()
            expect(p).to.be.a('promise')
            await p
        })

        it('should resolve to the model instance', async function() {
            const t = new Test
            expect(await t.delete()).to.be.equal(t)
        })

        it('should delete the record from the database', async function() {
            const rec = {
                name: 'test num a',
                num: 3.142,
                _id: new ObjectId
            }

            await scaffold.collection('TestModel').insertOne(rec)
            const t = new Test(rec)

            const lookup_1 = await scaffold.collection('TestModel').findOne({
                _id: rec._id
            })
            await t.delete()
            const lookup_2 = await scaffold.collection('TestModel').findOne({
                _id: rec._id
            })

            expect(lookup_1).to.be.eql(rec)
            expect(lookup_2).to.not.exist
        })
    })

    describe('the __collection method', function() {
        it('should be a function', function() {
            expect((new Test).__collection).to.be.a('function')
        })

        it('should return a mongodb collection', function() {
            const coll = (new Test).__collection()
            expect(coll).to.be.an.instanceof(Collection)
            expect(coll.collectionName).to.be.equal('TestModel')
        })
    })

    // TODO describe the __set_values method

    describe('the __scope_limit_save method', function() {
        it('should be a function', function() {
            expect((new Test).__scope_limit_save).to.be.a('function')
        })

        it('should return a promise', async function() {
            const t = new Test
            const p = t.__scope_limit_save({})
            expect(p).to.be.a('promise')
            await p
        })

        it('should pass the object through the model scopes', async function() {
            Test.scopes = [{save: sinon.fake((x) => x)}]
            const t = new Test
            const rec = {}
            const ret = await t.__scope_limit_save(rec)
            expect(ret).to.be.equal(rec)
            expect(Test.scopes[0].save.calledOnce).to.be.true
            expect(Test.scopes[0].save.getCall(0).args[0]).to.be.equal(rec)
        })
    })

    // --- Relations
    // TODO describe the has_one method
    // TODO describe the belongs_to_one method
    // TODO describe the has_many method
    // TODO describe the belongs_to_many method
})
