const { expect } = require('chai')

const Type = require('../../src/schema/Type')

describe('the schema types', function() {
    describe('the base type', function() {
        it('should be a functional class', function() {
            expect(Type).to.be.a('function')
        })

        it('should have a static validate method', function() {
            expect(Type.validate).to.be.a('function')
            expect(Type.validate.length).to.be.equal(1)
        })

        it('should have a static cast method', function () {
            expect(Type.cast).to.be.a('function')
            expect(Type.cast.length).to.be.equal(1)
        })
    })

    describe('the array schema type', function() {
        const ArrayType = require('../../src/schema/types/Array')

        it('should be a functional class', function() {
            expect(ArrayType).to.be.a('function')
        })

        it('should extend the base type', function () {
            expect(ArrayType.prototype).to.be.an.instanceof(Type)
        })

        it('should denote all types as valid', function() {
            const various_types = [
                'string',
                3.14,
                {},
                [],
                [true],
                true,
                () => {},
                undefined
            ]

            expect(various_types.every(x => ArrayType.validate(x)))
        })

        it('should cast items to arrays', function() {
            const test_arr = ['a', 'b', 'c']
            const test_val = 'a'

            expect(ArrayType.cast(test_arr)).to.have.all.members(test_arr)
            expect(ArrayType.cast(test_val)).to.be.an('array')
            expect(ArrayType.cast(test_val)).to.have.all.members([test_val])
        })
    })

    describe('the boolean schema type', function() {
        const BooleanType = require('../../src/schema/types/Boolean')

        it('should be a functional class', function() {
            expect(BooleanType).to.be.a('function')
        })

        it('should extend the base type', function () {
            expect(BooleanType.prototype).to.be.an.instanceof(Type)
        })

        it('should denote all types as valid', function() {
            const various_types = [
                'string',
                3.14,
                {},
                [],
                [true],
                true,
                () => {},
                undefined
            ]

            expect(various_types.every(x => BooleanType.validate(x)))
        })

        it('should cast items to boolean', function() {
            expect(BooleanType.cast('')).to.be.false
            expect(BooleanType.cast(true)).to.be.true
            expect(BooleanType.cast(false)).to.be.false
            expect(BooleanType.cast({})).to.be.true
            expect(BooleanType.cast([])).to.be.true
            expect(BooleanType.cast(0)).to.be.false
        })
    })

    describe('the date schema type', function() {
        const DateType = require('../../src/schema/types/Date')

        it('should be a functional class', function() {
            expect(DateType).to.be.a('function')
        })

        it('should extend the base type', function () {
            expect(DateType.prototype).to.be.an.instanceof(Type)
        })

        it('should validate Date instance and valid date strings', function() {
            expect(DateType.validate(new Date)).to.be.true
            expect(DateType.validate({})).to.be.false
            expect(DateType.validate('2020-02-02')).to.be.true
            expect(DateType.validate('Cannot be cast to date')).to.be.false
        })

        it('should cast items to dates', function() {
            expect(DateType.cast(new Date)).to.be.a('date')
            expect(DateType.cast('2020-02-02')).to.be.a('date')
        })
    })

    describe('the model schema type', function() {
        const ModelType = require('../../src/schema/types/Model')

        it('should be a functional class', function() {
            expect(ModelType).to.be.a('function')
        })

        it('should extend the base type', function () {
            expect(ModelType.prototype).to.be.an.instanceof(Type)
        })

        it('should validate for all objects', function() {
            expect(ModelType.validate({})).to.be.true
            expect(ModelType.validate(true)).to.be.false
            expect(ModelType.validate(() => {})).to.be.false
            expect(ModelType.validate({foo: true})).to.be.true
        })

        it('should not do any casting', function() {
            const various_types = [
                'string',
                3.14,
                {},
                [],
                [true],
                true,
                () => {},
                undefined
            ]

            expect(various_types.every(x => ModelType.cast(x) === x))
        })
    })

    describe('the number schema type', function() {
        const NumberType = require('../../src/schema/types/Number')

        it('should be a functional class', function() {
            expect(NumberType).to.be.a('function')
        })

        it('should extend the base type', function () {
            expect(NumberType.prototype).to.be.an.instanceof(Type)
        })

        it('should validate for all numerically castable things', function() {
            expect(NumberType.validate(3.14)).to.be.true
            expect(NumberType.validate('1004')).to.be.true
            expect(NumberType.validate('not a number')).to.be.false
            expect(NumberType.validate(NaN)).to.be.false
            expect(NumberType.validate({})).to.be.false
            expect(NumberType.validate(() => {})).to.be.false
            expect(NumberType.validate(true)).to.be.true
            expect(NumberType.validate(false)).to.be.true
        })

        it('should cast items to numbers', function() {
            expect(NumberType.cast(3.14)).to.be.equal(3.14)
            expect(NumberType.cast('1004')).to.be.equal(1004)
            expect(NumberType.cast(true)).to.be.equal(1)
            expect(NumberType.cast(false)).to.be.equal(0)
        })
    })

    describe('the object schema type', function() {
        const ObjectType = require('../../src/schema/types/Object')

        it('should be a functional class', function() {
            expect(ObjectType).to.be.a('function')
        })

        it('should extend the base type', function () {
            expect(ObjectType.prototype).to.be.an.instanceof(Type)
        })

        it('should validate for all serializable objects', function() {
            expect(ObjectType.validate({})).to.be.true
            expect(ObjectType.validate(() => {})).to.be.false
            expect(ObjectType.validate(3.14)).to.be.false
            expect(ObjectType.validate('test')).to.be.false
        })

        it('should cast deep copies of objects', function() {
            const base = {
                hello: 'world',
                foo: {
                    bar: 3.14
                }
            }

            const cast = ObjectType.cast(base)
            expect(cast).to.be.an('object')
            expect(cast).to.eql(base)

            cast.hello = 'there'
            expect(base.hello).to.be.equal('world')
        })
    })

    describe('the object id schema type', function() {
        const ObjectIdType = require('../../src/schema/types/ObjectId')
        const { ObjectId } = require('mongodb')

        it('should be a functional class', function() {
            expect(ObjectIdType).to.be.a('function')
        })

        it('should extend the base type', function () {
            expect(ObjectIdType.prototype).to.be.an.instanceof(Type)
        })

        it('should validate for all items castable to ObjectId', function() {
            expect(ObjectIdType.validate(new ObjectId)).to.be.true
            expect(ObjectIdType.validate(`${new ObjectId}`)).to.be.true
            expect(ObjectIdType.validate('Not a good id')).to.be.false
        })

        it('should cast items to ObjectId', function() {
            const base_id = new ObjectId
            const cast_1 = ObjectIdType.cast(base_id)
            const cast_2 = ObjectIdType.cast(`${base_id}`)

            expect(`${base_id}`).to.be.equal(`${cast_1}`)
            expect(`${base_id}`).to.be.equal(`${cast_2}`)
        })
    })

    describe('the string schema type', function() {
        const StringType = require('../../src/schema/types/String')

        it('should be a functional class', function() {
            expect(StringType).to.be.a('function')
        })

        it('should extend the base type', function () {
            expect(StringType.prototype).to.be.an.instanceof(Type)
        })

        it('should validate for all things castable to string', function() {
            expect(StringType.validate('hello')).to.be.true
            expect(StringType.validate(3.141)).to.be.true
            expect(StringType.validate({})).to.be.true
            expect(StringType.validate(() => {})).to.be.true
            expect(StringType.validate({ toString() { throw new Error } })).to.be.false
        })

        it('should cast items to string', function() {
            expect(StringType.cast('hello')).to.be.equal('hello')
            expect(StringType.cast(1440)).to.be.equal('1440')
            expect(StringType.cast({})).to.be.equal('[object Object]')
        })
    })
})
