const { expect } = require('chai')
const sinon = require('sinon')
const { DependencyInjector, Container } = require('flitter-di')
const { MongoClient } = require('mongodb')

const Connection = require('../../src/services/Connection')
const Collection = require('mongodb/lib/collection')

const common = require('../common')
const testing_mongo_url = common.test_connect

const get_setup = async (auto_open = true) => {
    const cont = new Container({ scaffold: Connection })
    const di = new DependencyInjector(cont)

    if ( auto_open ) await di.service('scaffold').open(testing_mongo_url)

    return { di }
}

describe('the scaffold service', function() {
    it('should be a functional class', function() {
        expect(Connection).to.be.a('function')
    })

    it('should extend flitter-di service', function() {
        expect(Connection.prototype).to.be.an.instanceof(require('flitter-di/src/Service'))
    })

    it('should be self-named scaffold', function() {
        expect(Connection.name).to.be.equal('scaffold')
    })

    describe('the close method', function() {
        it('should call the close method on the connection', async function() {
            const scaffold = new Connection()
            const fake_connection = {
                close: sinon.spy()
            }
            scaffold._connection = fake_connection

            await scaffold.close()
            expect(fake_connection.close.calledOnce).to.be.true
        })
    })

    describe('the open method', async function() {
        let di
        let conn
        beforeEach(async function() {
            const setup = await get_setup(false)
            di = setup.di
            conn = di.service('scaffold')
        })

        it('should be a function', function() {
            expect(conn.open).to.be.a('function')
        })

        it('should expect 1 argument', function() {
            expect(conn.open.length).to.be.equal(1)
        })

        it('should return a promise', async function() {
            const prom = conn.open(testing_mongo_url)
            expect(prom).to.be.a('promise')
            await prom
            await conn.close()
        })

        it('should open a connection', async function() {
            await conn.open(testing_mongo_url)
            expect(conn._connection).to.exist
            expect(conn._connection).to.be.an.instanceof(MongoClient)
            await conn.close()
        })

        it('should throw an error if the connection is invalid', async function() {
            let err
            try {
                await conn.open('bad_schema://example.com:27017/bad_url')
            } catch (e) {
                err = e
            }

            expect(err).to.exist
            expect(err).to.be.an('error')
        })
    })

    describe('the collection method', async function() {
        let di
        let conn
        beforeEach(async function() {
            const setup = await get_setup()
            di = setup.di
            conn = di.service('scaffold')
        })

        afterEach(async function() {
            await conn.close()
        })

        it('should be a function', function() {
            expect(conn.collection).to.be.a('function')
        })

        it('should expect one argument', function() {
            expect(conn.collection.length).to.be.equal(1)
        })

        it('should return the mongodb collection by name', function() {
            const coll = conn.collection('test_collection')
            expect(coll).to.be.an.instanceof(Collection)
            expect(coll.collectionName).to.be.equal('test_collection')
        })
    })
})
